"""Processors for FFT"""
import typing as t
from dataclasses import dataclass
from enum import Enum

import numpy as np

from .processor import Processor
from ..modules import Module, TorchModule

__all__ = ('FFTProcessor', 'FFTWrapperFunc', 'FFTType', 'FFTShift')


# Enumeration types
class FFTType(Enum):
    """The types of fft functions"""
    FFT = 'fft'  # Fast Fourier Transform (1D). exp(-i 2pi nu)
    IFFT = 'ifft'  # Inverse Fast Fourier Transform (1D). exp(i 2pi nu)
    RFFT = 'rfft'  # Real Fourier Transform (1D). Real component only.
    IRFFT = 'irfft'  # Inverse Real Fourier Transform (1D)
    NONE = None  # No Fourier Transform (null)


class FFTShift(Enum):
    """The type of post-FFT frequency/time shift to apply"""
    FFTSHIFT = 'fftshift'  # Shift zero-frequency to the center of the spectrum
    IFFTSHIFT = 'ifftshift'  # Reverse of FFTSHIFT
    NONE = None  # Do not apply frquency shift


# Wrapper Class Functions
@dataclass
class FFTWrapperFunc:
    """The fft wrapper callable object"""

    module: Module

    # Function parameter defaults that can be modified
    fft_type: FFTType = FFTType.FFT
    fft_shift: FFTShift = FFTShift.FFTSHIFT

    def __call__(self,
                 data: np.ndarray,
                 n: t.Optional[int] = None,
                 axis: int = -1,
                 norm: t.Optional[str] = None,
                 fft_type: t.Optional[FFTType] = None,
                 fft_shift: t.Optional[FFTShift] = None):
        """The fft wrapper function.

        Parameters
        ----------
        data
            The input data to Fourier transform
        n
            Length of the transformed axis of the output
        axis
            Axis over which to compute the FFT. (Last dimension used if not
            specified)
        norm
            Normalization mode. e.g. “backward”, “ortho”, “forward”
        fft_type
            The type of fft function to use. See :class:`.FFTType`
        fft_shift
            Center the zero-frequency to the center of the spectrum.
            See :class:`.FFTShift`

        Returns
        -------
        out
            The Fourier transormed and possibly frequeny
        """
        raise NotImplementedError


class NumpyFFTFunc(FFTWrapperFunc):
    """The numpy implementation of the FFT wrapper function"""

    def __call__(self, data, n=None, axis=-1, norm=None, fft_type=None,
                 fft_shift=None):
        # Setup arguments
        fft_type = fft_type if fft_type is not None else self.fft_type
        fft_shift = fft_shift if fft_shift is not None else self.fft_shift

        # Retrieve the 'fft' or 'ifft' function
        fft_func = self.module.get_callable(fft_type.value)

        # Perform the fft
        result = fft_func(a=data, n=n, axis=axis, norm=norm)
        result = result.astype(data.dtype)

        # Center the spectrum if needed
        if fft_shift is FFTShift.NONE:
            fftshift_func = None
        else:
            fftshift_func = self.module.get_callable(fft_shift.value)

        return fftshift_func(result) if fftshift_func is not None else result


class TorchFFTFunc(FFTWrapperFunc):
    """The Torch implementation of the FFT wrapper function"""

    from_numpy_module: Module

    def __call__(self, data, n=None, axis=-1, norm=None, fft_type=None,
                 fft_shift=None):
        # Setup arguments
        fft_type = fft_type if fft_type is not None else self.fft_type
        fft_shift = fft_shift if fft_shift is not None else self.fft_shift

        # Retrieve the conversion function for numpy arrays
        from_numpy = getattr(self.module.get_root_module(), 'from_numpy')

        # Convert the numpy ndarray to a tensor
        tensor = from_numpy(data)

        # Retrieve the 'fft' or 'ifft' function
        fft_func = self.module.get_callable(fft_type.value)

        # Perform the fft
        result = fft_func(input=tensor, n=n, dim=axis, norm=norm)

        # Center the spectrum if needed
        if fft_shift is FFTShift.NONE:
            fftshift_func = None
        else:
            fftshift_func = self.module.get_callable(fft_shift.value)

        return (fftshift_func(result).cpu().detach().numpy()
                if fftshift_func is not None else
                result.cpu().detach().numpy())


class FFTProcessor(Processor):
    """A processor with access to FFT functionality."""

    modules = (
        TorchModule('fft', 'torch.fft', TorchFFTFunc),
        Module('fft', 'numpy.fft', NumpyFFTFunc),
    )



"""
Setup CLI subcommand
"""
from inspect import isabstract
from itertools import groupby

import click
import yaml

from ..config import config as config_dict
from ..processors import Processor
from ..modules import Module
from ..utils.classes import all_subclasses


@click.group()
def setup():
    """Print information on the current setup"""
    pass


@setup.command()
def config():
    """Display and save the current configuration settings"""
    click.echo(yaml.safe_dump(config_dict))


@setup.command()
@click.option('--only-concrete',
              is_flag=True, default=False,
              help="Only show concrete (non-abstract) classes")
def processors(only_concrete):
    """List the available processors"""
    # Retrieve a list of all processor classes
    processor_clses = all_subclasses(Processor)

    # Retrieve only concrete classes--remove abstract classes--if specified
    if only_concrete:
        processor_clses = [processor_cls for processor_cls in processor_clses
                           if isabstract(processor_cls)]

    for count, processor_cls in enumerate(processor_clses, 1):
        click.echo(f"{count}. {processor_cls}")


@setup.command()
def modules():
    """List the available Python modules"""
    # Retrieve a list of all module instances
    module_objs = Module.list_instances()

    # Group by category
    modules_by_category = sorted(module_objs, key=lambda m: m.category)
    modules_by_category = groupby(modules_by_category, key=lambda m: m.category)

    for category, module_objs in modules_by_category:
        click.echo(click.style("Category: ", bold=True) + category)

        for count, module_obj in enumerate(module_objs, 1):
            module_obj.print()



"""
application configurations settings
"""
import yaml
from pathlib import Path
from importlib.metadata import entry_points  # python >=3.8
import typing as t
from types import ModuleType
from collections.abc import MutableMapping


class Settings(MutableMapping):
    """A lazy loading settings manager"""

    def add(self,
            config: t.Optional[dict] = None,
            module: t.Optional[ModuleType] = None,
            filepath: t.Optional[Path] = None):
        """Add a configuration data source."""

    def setup(self, **kwargs):
        pass


config = {
    'config': {
        'file': '~/.pocketchemistrc'
    },
    'cli': {
        'color': True
    },

}

# Implementation for loading configs from different places

def recursive_update(d, u):
    """Recursively update a dict."""
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d

# # load plugins configs
# for entrypoint in entry_points()['pocketchemist']:
#     if entrypoint.name == 'config':
#         try:
#             plugin_config = entrypoint.load()
#             recursive_update(config, plugin_config)
#         except ModuleNotFoundError:
#             logging.warning(f"The module '{entrypoint.value}' could not be "
#                             f"found.")
#
#
# # Load system config options
# for filepath in (Path.home() / config['config']['file'],):
#     if not filepath.exists():
#         continue
#         system_config = yaml.load(filepath)
#         recursive_update(config, system_config)



"""
Utilities for lists
"""
import typing as t


def wraplist(obj, default=None) -> t.List:
    """Wrap the object in a list, if needed

    Parameters
    ----------
    obj
        The object to wrap in a list
    default
        If obj is None, return the default instead. Not that if a default
        is not specified, the default returned object is an empty list.

    Examples
    --------
    >>> wraplist(None)
    []
    >>> wraplist('test')
    ['test']
    >>> wraplist([1, 2, 3])
    [1, 2, 3]
    """
    if obj is None:
        return default if default is not None else []
    elif hasattr(obj, '__iter__') and not isinstance(obj, str):
        # Nothing to do, we've good
        return obj
    else:
        # wrap in a list
        return [obj]



"""Classes and functions to conduct one or more processes"""
from .processor import Processor, GroupProcessor
from .fft import FFTProcessor, FFTType

__all__ = ('Processor', 'GroupProcessor', 'FFTProcessor', 'FFTType')



"""
Utilities for classes
"""


def all_subclasses(cls):
    """Retrieve all subclasses, sub-subclasses and so on for a class
    Parameters
    ----------
    cls : Type
        The class object to inspect for subclasses.
    Returns
    -------
    subclasses : list
        The list of all subclasses.
    Examples
    --------
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(B): pass
    >>> all_subclasses(A)
    [<class 'disseminate.utils.classes.B'>, \
<class 'disseminate.utils.classes.C'>]
    """
    return cls.__subclasses__() + [g for s in cls.__subclasses__()
                                   for g in all_subclasses(s)]



"""
Classes for managing module imports and availability information.
"""
import abc
from dataclasses import dataclass
from importlib import import_module, metadata  # metadata in python >= 3.8
from types import ModuleType
import typing as t

import click

__all__ = ('Module', 'GPUModule', 'TorchModule')

# The number of spaces to add for printing sub-levels of processors
space_level = 2


@dataclass
class Module:

    #: The category name of the module
    category: str

    #: The name of the module
    name: str

    #: The name of the imported function (str), a wrapper class that needs
    #: to be instantiated or directly a callable
    callable: t.Union[str, t.Callable]

    #: Options for printing the module to the terminal
    print_cls_fg_color = 'cyan'

    _modules = dict()

    def get_module(self) -> t.Union[ModuleType, None]:
        """Retrieve the module associated with this object

        Returns
        -------
        module
            The loaded module, if available
            None if the module could not be found
        """
        key = self.name

        if key in Module._modules:
            return Module._modules[key]

        # Try loading the module
        try:
            module = import_module(key)
            Module._modules[key] = module
        except ImportError:
            Module._modules[key] = None

        return Module._modules[key]

    def get_root_module(self) -> t.Union[ModuleType, None]:
        """Retrieve the root module of package

        Returns
        -------
        module
            The loaded module, if available
            None if the module could not be found
        """
        # Try to get the root module's name
        root_module_name = self.name.split('.')[0]

        # Try loading the module
        try:
            return import_module(root_module_name)
        except ImportError:
            return None

    def get_callable(self,
                     callable_obj: t.Optional[t.Union[str, t.Callable,
                                                      type]] = None) \
            -> t.Union[t.Callable, None]:
        """Retrieve the module function.

        Parameters
        ----------
        callable_obj
            Optional callable object that is used instead of self.callable.
            This can be a string for a callable (function) in the module,
            a callable directly, like a function or class, or class to generate
            a callable, which takes a single 'module' parameter, which is this
            module.

        Returns
        -------
        callable
            The module's function or callable class, if available.
            None if the function could not be found.
        """
        # Setup the arguments
        callable_obj = (callable_obj if callable_obj is not None else
                        self.callable)

        # See if the module is available, and retrieve the function if it is
        module = self.get_module()
        if isinstance(callable_obj, str):
            # The callable is a string, which should be retrieved from the
            # module
            return getattr(module, callable_obj, None)
        elif type(callable_obj) == type:
            # The callable is a wrapper class that should be instantiated
            # with this module
            return callable_obj(self)
        else:
            # Otherwise return the callable itself
            assert callable(callable_obj)
            return callable_obj

    @classmethod
    def list_instances(cls):
        """List modules managed by this class"""
        import gc
        return [obj for obj in gc.get_objects() if isinstance(obj, cls)]

    def print(self,
              space_level: int = space_level,
              item_number: t.Optional[str] = "",
              subitems: t.Optional[t.List[str]] = None) -> None:
        """Print information on the module to the terminal.

        Parameters
        ----------
        space_level
            The number of spaces to separate a level
        item_number
            An optional character to prepend the printed processing string.
        subitems
            Additional items to print about the module as subitems.
        """
        cls_name = self.__class__.__name__
        item_number = str(item_number) + '. ' if item_number else ""
        subitems = list(subitems) if subitems is not None else []

        # Setup subitems to print
        module = self.get_module()
        module_available = module is not None
        callable_available = self.get_callable() is not None
        subitems[:0] += [click.style("module: ") +
                     (click.style("AVAILABLE", fg='green') if module_available
                      else click.style("NOT FOUND", fg='red')) + " " +
                     click.style("callable: ") +
                     (click.style("AVAILABLE", fg='green') if callable_available
                      else click.style("NOT FOUND", fg='red'))]

        # Add version information
        if module_available:
            # Find the root module
            root_module = self.get_root_module()
            if hasattr(root_module, '__version__'):
                subitems += [click.style("version: ") +
                             click.style(root_module.__version__)]

        # Format the string to print
        click.echo(" " * space_level +
                   item_number +
                   click.style(cls_name, fg=self.print_cls_fg_color) +
                   f"({self.name})")
        for subitem in subitems:
            click.echo(" " * space_level * 2 + subitem)


class GPUModule(Module):

    @abc.abstractmethod
    def get_available(self):
        """Determine if the GPU libraries are available"""
        raise NotImplementedError


class TorchModule(GPUModule):
    """A module for PyTorch"""

    def gpu_available(self):
        root_module = self.get_root_module()
        try:
            return root_module.cuda.is_available()
        except AttributeError:
            return False

    def print(self,
              space_level: int = space_level,
              item_number: t.Optional[str] = "",
              subitems: t.Optional[t.List[str]] = None) -> None:
        # Setup arguments
        subitems = subitems if subitems is not None else []

        gpu_available = self.gpu_available()
        subitems += [click.style("GPU (CUDA) Available: ") +
                     (click.styles(gpu_available, fg='green')
                      if gpu_available else
                      click.style(gpu_available, fg='red'))]

        super().print(space_level=space_level, item_number=item_number,
                      subitems=subitems)



from importlib.metadata import entry_points  # python >=3.8
import sys

import click
from loguru import logger

from .setup import setup


@click.group()
@click.option('--debug', is_flag=True, default=False,
              help="Display debugging information")
def main(debug):
    """A pocket chemist to analyze spectra and molecules"""
    # Remove default logger
    logger.remove()

    # Configure logger and set default level
    if debug:
        logger.add(sys.stderr, level="DEBUG", enqueue=True)
        logger.debug("Debug mode ON")
    else:
        logger.add(sys.stderr, level="WARNING", enqueue=True)


# Add subcommands
main.add_command(setup)


# load plugins
for entrypoint in entry_points().get('pocketchemist', []):
    if entrypoint.name == 'cli':
        try:
            main.add_command(entrypoint.load())
        except ModuleNotFoundError:
            logger.warning(f"The module '{entrypoint.value}' could not be "
                           f"found.")




# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pocketchemist',
 'pocketchemist.cli',
 'pocketchemist.modules',
 'pocketchemist.processors',
 'pocketchemist.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.0.3,<9.0.0',
 'loguru>=0.5.3,<0.6.0',
 'numpy>=1.22.0,<2.0.0',
 'scipy>=1.7.3,<2.0.0']

extras_require = \
{'torch': ['torch>=1.10,<2.0']}

entry_points = \
{'console_scripts': ['pc = pocketchemist.cli:main']}

setup_kwargs = {
    'name': 'pocketchemist',
    'version': '0.1.1',
    'description': 'Software for the analysis of spectra and molecules',
    'long_description': None,
    'author': 'J Lorieau',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
Processor base classes
"""
import abc
import typing as t

import click
from pocketchemist.utils.list import wraplist

from ..modules import Module

__all__ = ('Processor', 'GroupProcessor')

# The number of spaces to add for printing sub-levels of processors
space_level = 2


class ProcessorMeta(type):
    def __repr__(cls):
        """A string representation for Processor classes"""
        return f"{cls.__name__}(module='{cls.__module__}')"


class Processor(metaclass=ProcessorMeta):
    """A processor for data and other objects"""

    #: The name of the processor
    name: str = None

    #: Parameters used for processing
    params: t.Dict[str, t.Union[str, int, float, 'lmfit.Parameter']] = None

    #: A listing of required and optional parameters
    required_params: tuple = ()
    optional_params: tuple = ()

    #: A list of modules the processor depends on
    modules = t.Iterable[Module]

    #: Options for printing the processor to the terminal
    print_cls_fg_color = 'cyan'

    def __init__(self, name=None, *args, **kwargs):
        # Set the processor name
        if name is not None:
            self.name = name

        # Setup other attributes
        self.params = dict()

        # Set the parameters
        # First check if there are missing parameters
        missing_params = set(self.required_params) - kwargs.keys()
        if len(missing_params) > 0:
            raise ValueError(f"The {self.__class__.__name__} processor is "
                             f"missing the following required kwargs "
                             f"parameters: {missing_params}")

        # Collect the specified parameters
        available_params = (set(self.required_params) |
                            set(self.optional_params))
        specified_params = available_params.intersection(kwargs.keys())
        for specified_param in specified_params:
            self.params[specified_param] = kwargs[specified_param]

    def __repr__(self):
        name = self.name if self.name is not None else self.__class__.__name__
        return f"{name}"

    def __getattr__(self, name):
        """Search the params dict for an attribute"""
        if isinstance(self.params, dict) and name in self.params:
            return self.params.get(name)
        raise AttributeError(f"'{self.__class__.__name__}' object has no "
                             f"attribute '{name}'")

    @abc.abstractmethod
    def process(self, **kwargs: dict):
        """Conduct the processing

        Parameters
        ----------
        kwargs
            The input and output parameters for the process. Processed data
            should be placed in the kwargs for subsequent processors
        """
        pass

    @classmethod
    def get_module_callable(cls,
                            category: t.Optional[str] = None,
                            name: t.Optional[str] = None,
                            modules: t.Optional[t.Iterable[Module]] = None) \
            -> t.Callable:
        """Retrieve the first (best) module callable

        Parameters
        ----------
        category
            If specified, filter modules by the given category.
        name
            If specified, search for the callable with the given module name.
            Otherwise, the first module found will be used.
        modules
            The list of modules (:obj:`Module`) to search for the callable.
            If the list of modules is not specified, the self.modules list
            will be used.

        Returns
        -------
        callable
            The found module and its callable (function).

        Raises
        ------
        ModuleNotFoundError
            Raised when a suitable module could not be found.
        """
        # Setup the list of modules to search
        modules = wraplist(modules, default=getattr(cls, 'modules', []))

        # Filter the list, if needed
        if category is not None:
            modules = [module for module in modules
                       if module.category == category]
        if name is not None:
            modules = [module for module in modules if module.name == name]

        # Search for the callable
        for module in modules:
            callable_func = module.get_callable()
            if callable_func is not None:
                return callable_func

        # None found! We have a problem and should raise an exception
        if name is not None:
            msg = f"Module with name '{name}' could not be found"
        else:
            msg = "A suitable module was not found"
        raise ModuleNotFoundError(msg)

    def print(self, level: int = 0, space_level: int = 0,
              item_number: t.Optional[str] = "") -> None:
        """Print information on the processor to the terminal.

        Parameters
        ----------
        level
            The level of the processor for processors that are nested in
            groups.
        space_level
            The number of spaces to separate a level
        item_number
            An optional character to prepend the printed processing string.
        """
        name = self.name if self.name is not None else self.__class__.__name__
        item_number = str(item_number) + '. ' if item_number else ""
        params = ", ".join("=".join((k, v)) for k, v in self.params.items())
        params = f"({params})" if params else ""

        click.echo(" " * (level * space_level) +
                   item_number +
                   click.style(f"{name}", fg=self.print_cls_fg_color) +
                   params, err=True)


class GroupProcessor(Processor):
    """A processor comprising multiple subprocessors"""

    #: A list of (sub-)processors owned by this group
    processors: t.List[Processor] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.processors = []

    def __repr__(self):
        name = self.name if self.name is not None else self.__class__.__name__
        num_processors = len(self.processors)
        return f"{name}(number_processors={num_processors})"

    def __iadd__(self, other):
        self.add(other)
        return self

    def add(self, processor: Processor):
        """Add a processor to the list of processors

        Parameters
        ----------
        processor
            The processor to add to the self.processors list
        """
        if isinstance(processor, Processor):
            self.processors.append(processor)
        else:
            raise TypeError(f"The {self.__class__.__name__} class can only "
                            f"append Processor type objects.")

    def process(self, **kwargs):
        """Conduct the group processing"""
        if not hasattr(self.processors, '__iter__'):
            return None

        # Iterate and process the processors
        for processor in self.processors:
            rv = processor.process(**kwargs)

    def print(self, level=0, space_level=space_level):
        """Print information on the processor to the terminal"""
        super().print(level=level, space_level=space_level)

        for count, processor in enumerate(self.processors, 1):
            processor.print(level=level + 1, space_level=space_level,
                            item_number=count)

